from coalib.core.Graphs import traverse_graph


class DependencyTracker:
    """
    A ``DependencyTracker`` allows to register and manage dependencies between
    objects.

    This class uses a directed graph to track relations.

    Add a dependency relation between two objects:

    >>> object1 = object()
    >>> object2 = object()
    >>> tracker = DependencyTracker()
    >>> tracker.add(object2, object1)

    This would define that ``object1`` is dependent on ``object2``.

    If you define that ``object2`` has its dependency duty fulfilled, you can
    resolve it:

    >>> resolved = tracker.resolve(object2)
    >>> resolved
    {<object object at ...>}
    >>> resolved_object = resolved.pop()
    >>> resolved_object is object1
    True

    This returns all objects that are now freed, meaning they have no
    dependencies any more.

    >>> object3 = object()
    >>> tracker.add(object2, object1)
    >>> tracker.add(object3, object1)
    >>> tracker.resolve(object2)
    set()
    >>> tracker.resolve(object3)
    {<object object at ...>}

    The ones who instantiate a ``DependencyTracker`` are responsible for
    resolving dependencies in the right order. Dependencies which are itself
    dependent will be forcefully resolved and removed from their according
    dependencies too.
    """

    def __init__(self):
        self._dependency_dict = {}

    def get_dependants(self, dependency):
        """
        Returns all immediate dependants for the given dependency.

        >>> tracker = DependencyTracker()
        >>> tracker.add(0, 1)
        >>> tracker.add(0, 2)
        >>> tracker.add(1, 3)
        >>> tracker.get_dependants(0)
        {1, 2}
        >>> tracker.get_dependants(1)
        {3}
        >>> tracker.get_dependants(2)
        set()

        :param dependency:
            The dependency to retrieve all dependants from.
        :return:
            A set of dependants.
        """
        try:
            return set(self._dependency_dict[dependency])
        except KeyError:
            return set()

    def get_dependencies(self, dependant):
        """
        Returns all immediate dependencies of a given dependant.

        >>> tracker = DependencyTracker()
        >>> tracker.add(0, 1)
        >>> tracker.add(0, 2)
        >>> tracker.add(1, 2)
        >>> tracker.get_dependencies(0)
        set()
        >>> tracker.get_dependencies(1)
        {0}
        >>> tracker.get_dependencies(2)
        {0, 1}

        :param dependant:
            The dependant to retrieve all dependencies from.
        :return:
            A set of dependencies.
        """
        return set(
            dependency
            for dependency, dependants in self._dependency_dict.items()
            if dependant in dependants)

    def get_all_dependants(self, dependency):
        """
        Returns a set of all dependants of the given dependency, even
        indirectly related ones.

        >>> tracker = DependencyTracker()
        >>> tracker.add(0, 1)
        >>> tracker.add(1, 2)
        >>> tracker.get_all_dependants(0)
        {1, 2}

        :param dependency:
            The dependency to get all dependants for.
        :return:
            A set of dependants.
        """
        dependants = set()

        def append_to_dependants(prev, nxt):
            dependants.add(nxt)

        traverse_graph(
            [dependency],
            lambda node: self._dependency_dict.get(node, frozenset()),
            append_to_dependants)

        return dependants

    def get_all_dependencies(self, dependant):
        """
        Returns a set of all dependencies of the given dependants, even
        indirectly related ones.

        >>> tracker = DependencyTracker()
        >>> tracker.add(0, 1)
        >>> tracker.add(1, 2)
        >>> tracker.get_all_dependencies(2)
        {0, 1}

        :param dependant:
            The dependant to get all dependencies for.
        :return:
            A set of dependencies.
        """
        dependencies = set()

        def append_to_dependencies(prev, nxt):
            dependencies.add(nxt)

        traverse_graph(
            [dependant],
            lambda node:
                {dependency
                 for dependency, dependants in self._dependency_dict.items()
                 if node in dependants},
            append_to_dependencies)

        return dependencies

    def add(self, dependency, dependant):
        """
        Add a dependency relation.

        This function does not check for circular dependencies.

        >>> tracker = DependencyTracker()
        >>> tracker.add(0, 1)
        >>> tracker.add(0, 2)
        >>> tracker.resolve(0)
        {1, 2}

        :param dependency:
            The object that is the dependency.
        :param dependant:
            The object that is the dependant.
        """
        if dependency not in self._dependency_dict:
            self._dependency_dict[dependency] = set()

        self._dependency_dict[dependency].add(dependant)

    def resolve(self, dependency):
        """
        Resolves all dependency-relations from the given dependency, and frees
        and returns dependants with no more dependencies. If the given
        dependency is itself a dependant, all those relations are also removed.

        >>> tracker = DependencyTracker()
        >>> tracker.add(0, 1)
        >>> tracker.add(0, 2)
        >>> tracker.add(2, 3)
        >>> tracker.resolve(0)
        {1, 2}
        >>> tracker.resolve(2)
        {3}
        >>> tracker.resolve(2)
        set()

        :param dependency:
            The dependency.
        :return:
            Returns a set of dependants whose dependencies were all resolved.
        """
        # Check if dependency has itself dependencies which aren't resolved,
        # these need to be removed too. This operation does not free any
        # dependencies.
        dependencies_to_remove = []
        for tracked_dependency, dependants in self._dependency_dict.items():
            if dependency in dependants:
                dependants.remove(dependency)

                # If dependants set is now empty, schedule dependency for
                # removal from dependency_dict.
                if not dependants:
                    dependencies_to_remove.append(tracked_dependency)

        for tracked_dependency in dependencies_to_remove:
            del self._dependency_dict[tracked_dependency]

        # Now free dependants which do depend on the given dependency.
        possible_freed_dependants = self._dependency_dict.pop(
            dependency, set())
        non_free_dependants = set()

        for possible_freed_dependant in possible_freed_dependants:
            # Check if all dependencies of dependants from above are satisfied.
            # If so, there are no more dependencies for dependant. Thus it's
            # resolved.
            for dependants in self._dependency_dict.values():
                if possible_freed_dependant in dependants:
                    non_free_dependants.add(possible_freed_dependant)
                    break

        # Remaining dependents are officially resolved.
        return possible_freed_dependants - non_free_dependants

    def check_circular_dependencies(self):
        """
        Checks whether there are circular dependency conflicts.

        >>> tracker = DependencyTracker()
        >>> tracker.add(0, 1)
        >>> tracker.add(1, 0)
        >>> tracker.check_circular_dependencies()
        Traceback (most recent call last):
         ...
        coalib.core.CircularDependencyError.CircularDependencyError: ...

        :raises CircularDependencyError:
            Raised on circular dependency conflicts.
        """
        traverse_graph(
            self._dependency_dict.keys(),
            lambda node: self._dependency_dict.get(node, frozenset()))

    @property
    def are_dependencies_resolved(self):
        """
        Checks whether all dependencies in this ``DependencyTracker`` instance
        are resolved.

        >>> tracker = DependencyTracker()
        >>> tracker.are_dependencies_resolved
        True
        >>> tracker.add(0, 1)
        >>> tracker.are_dependencies_resolved
        False
        >>> tracker.resolve(0)
        {1}
        >>> tracker.are_dependencies_resolved
        True

        :return:
            ``True`` when all dependencies resolved, ``False`` if not.
        """
        return not self._dependency_dict



import argparse
import datetime
from distutils.core import Command
from distutils.errors import DistutilsOptionError


class BuildManPage(Command):
    """
    Add a ``build_manpage`` command  to your setup.py.
    To use this Command class add a command to call this class::

        # For setuptools
        setup(
              entry_points={
                "distutils.commands": [
                    "build_manpage = coalib.misc.BuildManPage:BuildManPage"
                ]
              }
        )

        # For distutils
        from coalib.misc.BuildManPage import BuildManPage
        setup(
              cmdclass={'build_manpage': BuildManPage}
        )

    You can then use the following setup command to produce a man page::

        $ python setup.py build_manpage --output=coala.1 \
            --parser=coalib.parsing.DefaultArgParser:default_arg_parser

    If automatically want to build the man page every time you invoke
    your build, add to your ``setup.cfg`` the following::

        [build_manpage]
        output = <appname>.1
        parser = <path_to_your_parser>
    """
    user_options = [
        ('output=', 'O', 'output file'),
        ('parser=', None, 'module path to an ArgumentParser instance'
         '(e.g. mymod:func, where func is a method or function which return'
         'an arparse.ArgumentParser instance.'),
    ]

    def initialize_options(self):
        self.output = None
        self.parser = None

    def finalize_options(self):
        if self.output is None:
            raise DistutilsOptionError('\'output\' option is required')
        if self.parser is None:
            raise DistutilsOptionError('\'parser\' option is required')
        mod_name, func_name = self.parser.split(':')
        fromlist = mod_name.split('.')
        mod = __import__(mod_name, fromlist=fromlist)
        self._parser = (
            getattr(mod, func_name)(formatter_class=ManPageFormatter))

        self.announce('Writing man page %s' % self.output)
        self._today = datetime.date.today()

    def run(self):
        dist = self.distribution
        homepage = dist.get_url()
        maintainer = dist.get_maintainer()
        _license = dist.get_license()
        appname = self._parser.prog

        sections = {'see also': ('Online documentation: {}'.format(homepage)),
                    'maintainer(s)': maintainer,
                    'license': _license}

        dist = self.distribution
        mpf = ManPageFormatter(appname,
                               desc=dist.get_description(),
                               long_desc=dist.get_long_description(),
                               ext_sections=sections,
                               parser=self._parser)

        formatted_man_page = mpf.format_man_page()

        with open(self.output, 'w') as man_file:
            man_file.write(formatted_man_page)


class ManPageFormatter(argparse.HelpFormatter):

    def __init__(self,
                 prog,
                 indent_increment=2,
                 max_help_position=24,
                 width=None,
                 desc=None,
                 long_desc=None,
                 ext_sections=None,
                 parser=None):
        argparse.HelpFormatter.__init__(self, prog)

        self._prog = prog
        self._section = 1
        self._today = datetime.date.today().strftime('%Y\\-%m\\-%d')
        self._desc = desc
        self._long_desc = long_desc
        self._ext_sections = ext_sections
        self._parser = parser

    def _format_action_invocation(self, action):
        if not action.option_strings:
            metavar, = self._metavar_formatter(action, action.dest)(1)
            return metavar

        else:
            # if the Optional doesn't take a value, format is:
            #    -s, --long
            if action.nargs == 0:
                parts = [ManPageFormatter._bold(action_str)
                         for action_str in action.option_strings]

            # if the Optional takes a value, format is:
            #    -s ARGS, --long ARGS
            else:
                default = ManPageFormatter._underline(action.dest.upper())
                args_string = self._format_args(action, default)
                parts = ['%s %s' % (self._bold(option_string), args_string)
                         for option_string in action.option_strings]

            return ', '.join(parts)

    @staticmethod
    def _markup(string):
        return string.replace('-', '\\-')

    @staticmethod
    def _add_format(string, front, back):
        if not string.strip().startswith(front):
            string = front + string
        if not string.strip().endswith(back):
            string = string + back
        return string

    @staticmethod
    def _underline(string):
        return ManPageFormatter._add_format(string, '\\fI', '\\fR')

    @staticmethod
    def _bold(string):
        return ManPageFormatter._add_format(string, '\\fB', '\\fR')

    def _mk_title(self):
        return '.TH {0} {1} {2}\n'.format(self._prog,
                                          self._section,
                                          self._today)

    def _mk_name(self):
        return '.SH NAME\n%s\n' % (self._parser.prog)

    def _mk_synopsis(self):
        self.add_usage(self._parser.usage,
                       self._parser._actions,
                       self._parser._mutually_exclusive_groups,
                       prefix='')
        usage = self._format_usage(None,
                                   self._parser._actions,
                                   self._parser._mutually_exclusive_groups,
                                   '')

        usage = usage.replace('%s ' % self._prog, '')
        usage = ('.SH SYNOPSIS\n \\fB%s\\fR %s\n'
                 % (ManPageFormatter._markup(self._prog), usage))
        return usage

    def _mk_description(self):
        if self._long_desc:
            long_desc = self._long_desc.replace('\n', '\n.br\n')
            return '.SH DESCRIPTION\n%s\n' % self._markup(long_desc)
        else:
            return ''

    def _mk_options(self):
        formatter = self._parser._get_formatter()

        # positionals, optionals and user-defined groups
        for action_group in self._parser._action_groups:
            formatter.start_section(None)
            formatter.add_text(None)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()

        # epilog
        formatter.add_text(self._parser.epilog)

        # determine help from format above
        return '.SH OPTIONS\n' + formatter.format_help()

    def _mk_footer(self):
        sections = self._ext_sections
        if not hasattr(sections, '__iter__'):
            return ''

        footer = []

        for section in sorted(sections.keys()):
            part = '.SH {}\n {}'.format(section.upper(), sections[section])
            footer.append(part)

        return '\n'.join(footer)

    def format_man_page(self):
        page = []
        page.append(self._mk_title())
        page.append(self._mk_name())
        page.append(self._mk_synopsis())
        page.append(self._mk_description())
        page.append(self._mk_options())
        page.append(self._mk_footer())

        return ''.join(page)



from coalib.core.CircularDependencyError import CircularDependencyError


def traverse_graph(start_nodes, get_successive_nodes,
                   run_on_edge=lambda prev, nxt: None):
    """
    Traverses all edges of a directed, possibly disconnected graph once.
    Detects cyclic graphs by raising a ``CircularDependencyError``.

    >>> graph = {1: [2], 2: [3, 4], 5: [3], 3: [6]}
    >>> def get_successive_nodes(node):
    ...     return graph.get(node, [])
    >>> edges = set()
    >>> def append_to_edges(prev, nxt):
    ...     edges.add((prev, nxt))
    >>> traverse_graph([1, 5], get_successive_nodes, append_to_edges)
    >>> sorted(edges)
    [(1, 2), (2, 3), (2, 4), (3, 6), (5, 3)]

    You can also use this function to detect cyclic graphs:

    >>> graph = {1: [2], 2: [3], 3: [1]}
    >>> traverse_graph([1], get_successive_nodes)
    Traceback (most recent call last):
     ...
    coalib.core.CircularDependencyError.CircularDependencyError: ...

    :param start_nodes:
        The nodes where to start traversing the graph.
    :param get_successive_nodes:
        A callable that takes in a node and returns an iterable of nodes to
        traverse next.
    :param run_on_edge:
        A callable that is run on each edge during traversing. Takes in two
        parameters, the previous- and next-node which form an edge. The default
        is an empty function.
    :raises CircularDependencyError:
        Raised when the graph is cyclic.
    """
    path = set()
    visited_nodes = set()

    def visit(node):
        if node not in visited_nodes:
            visited_nodes.add(node)
            path.add(node)

            for subnode in get_successive_nodes(node):
                run_on_edge(node, subnode)

                if subnode in path:
                    raise CircularDependencyError(subnode)

                visit(subnode)

            path.remove(node)

    for node in start_nodes:
        visit(node)
